import 'dart:convert';

import 'package:contentsquare/contentsquare.dart';
import 'package:contentsquare/src/modules/logger/logger.dart';
import 'package:contentsquare/src/modules/screengraph/compression.dart';
import 'package:contentsquare/src/modules/screengraph/render_object_extension.dart';
import 'package:contentsquare/src/modules/webview_tracking/javascript_event_handler.dart';
import 'package:fast_immutable_collections/fast_immutable_collections.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';

import '../screengraph/hit_test/get_element_data_of_hit_render_object.dart';
import '../screengraph/hit_test/hit_element_data.dart';
import 'exceptions/exceptions.dart';
import 'logs/logs.dart';
import 'screengraph_retrieval.dart';

/// Widget used to track a WebView.
///
/// {@template contentsquare.webview.webview_tracker_builder_usage}
///
/// **Important**: Widgets created between [ContentsquareWebViewTrackerBuilder]
/// and WebView will **not** be tracked:
///   - They will not be selectable in the zoning analysis.
///   - Gestures on those widgets will not be tracked.
/// Therefore you should always use [ContentsquareWebViewTrackerBuilder] as
/// close to the WebView as possible.
/// {@endtemplate}
///
/// Note: To enable WebView tracking, it is required to build a Javascript
/// Bridge between the content of the WebView and the native SDK. To do so,
/// you will have to implement the Contentsquare WebView Javascript Tracking
/// Tag in the web pages called in your app WebViews.
///   - See https://docs.contentsquare.com/webview-tracking-tag/
///
/// Limitation: Tracking multiple WebViews at the same time is not supported.
///
/// {@endtemplate}
class ContentsquareWebViewTrackerBuilder extends StatefulWidget {
  /// {@template contentsquare.webview.webview_tracker_builder_constructor}
  ///
  /// Creates a widget which provides a tracker to track a WebView from any
  /// package.
  ///
  /// For the tracking to work, you need to:
  ///   - Enable javascript in the WebView.
  ///   - Register the given [ContentsquareWebViewTracker.javascriptChannels]
  ///     to allow the webview to communicate with the SDK.
  ///   - Use [ContentsquareWebViewTracker.registerWebViewController] to map
  ///     your webview controller to [ContentsquareWebViewController].
  ///     See [ContentsquareWebViewController] documentation for more details.
  ///   - Use [ContentsquareWebViewTracker.startPageTracking] to start
  ///     the tracking.
  ///     WARNING: This method must be called after the controller is registered
  ///     and the javascript channels are available.
  ///
  /// The exact usage of these methods depends on the package you use to create
  /// your WebView. Reach out to us if you need help.
  ///
  /// {@endtemplate}
  const ContentsquareWebViewTrackerBuilder({
    Key? key,
    required this.builder,
  }) : super(key: key);

  /// {@template contentsquare.webview.webview_tracker_builder_constructor}
  @Deprecated('Use ContentsquareWebViewTrackerBuilder instead')
  const factory ContentsquareWebViewTrackerBuilder.custom({
    Key? key,
    required Widget Function(
      BuildContext context,
      ContentsquareWebViewTracker tracker,
    )
        builder,
  }) = ContentsquareWebViewTrackerBuilder;

  /// Callback in which the WebView should be built.
  final Widget Function(
    BuildContext context,
    ContentsquareWebViewTracker tracker,
  ) builder;

  @override
  State<ContentsquareWebViewTrackerBuilder> createState() =>
      _a134();
}


class _a134
    extends State<ContentsquareWebViewTrackerBuilder> {
  
  late final a75 = a137(
    a72: a334,
    a71: _a216,
  );

  
  String? _a216() {
    
    
    
    
    if (!mounted) {
      a187.a39(a177());
      return null;
    }

    
    
    
    
    
    
    final a308 = context as Element;

    
    
    
    
    
    
    
    
    
    
    final a309 = a308.renderObject;
    if (a309 == null) {
      a187.a39(a179());
      return null;
    }
    if (a309 is! a136) {
      a187.a35(
        'The type of render object associated with the $ContentsquareWebViewTrackerBuilder widget is ${a309.runtimeType}.\n'
        'This is an error, it should always be a $a136.\n',
      );
      return null;
    }

    
    
    
    
    
    
    
    
    final a310 = a309.a226;
    if (a310 == null) {
      a187.a39(a178());
      return null;
    }

    final a311 = a263(
      a524: a309,
      a525: a140.a232(
        a494: a310.center,
      ),
    );

    return a311.a13(
      a422: (hitElementData) {
        return a193(hitElementData.a67);
      },
      
      
      a421: () => null,
    );
  }

  @override
  void dispose() {
    a75._a223();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return _a138(
      a487: _a135(
        a482: widget.builder(context, a75),
        a74: a75.a23,
      ),
    );
  }
}


class _a135
    extends SingleChildRenderObjectWidget {
  const _a135({
    Key? a481,
    required this.a74,
    required Widget a482,
  }) : super(key: a481, child: a482);

  
  final Future<Map<String, dynamic>> Function() a74;

  @override
  RenderObject createRenderObject(BuildContext context) {
    return a136(a73: a74);
  }
}


class a136 extends RenderProxyBox {
  
  a136({
    required this.a73,
  });

  
  final Future<Map<String, dynamic>> Function() a73;
}


class a137 implements ContentsquareWebViewTracker {
  
  a137({
    required this.a72,
    required this.a71,
  });

  
  final a142 a72;

  
  final String? Function() a71;

  
  ContentsquareWebViewController? _a312;

  @override
  void registerWebViewController(ContentsquareWebViewController controller) {
    _a312 = controller;
  }

  @override
  Future<void> startPageTracking() async {
    final _a313 = this._a312;

    if (_a313 == null) {
      return a187.a39(
        a156(
          a521: null,
          a522: 'WebViewController is null, did you forget to use '
              'tracker.onWebViewCreated?',
        ),
      );
    }

    
    final a314 = _a313.isChannelAvailable;
    if (a314 != null) {
      final bool a315;
      try {
        a315 =
            await _a218(
          a484: a314,
        );
      } catch (error) {
        return a187.a39(
          a156(
            a521: await _a313.currentUrl?.call(),
            a522: 'Unexpected error while executing JS code, check the '
                'registration of the javascript channels: $error',
          ),
        );
      }
      if (!a315) {
        return a187.a39(
          a156(
            a521: await _a313.currentUrl?.call(),
            a522:
                'JavascriptChannels are not registered, did you forget to use '
                'tracker.javascriptChannels?',
          ),
        );
      }
    }

    try {
      await _a313.evaluateJavascript(
        '''
${_a217(a483: _a313.javascriptChannelCall)}
        
// This is needed because [_webViewController.evaluateJavascript] only works on
// js code that returns a value (and the bridge registration does not return a 
// value)
true;
''',
      );
    } catch (error) {
      return a187.a39(
        a156(
          a521: await _a313.currentUrl?.call(),
          a522: 'Unexpected error while executing JS code to setup the '
              'Javascript bridge: $error',
        ),
      );
    }

    return a187.a39(
      a158(
        a523: await _a313.currentUrl?.call(),
      ),
    );
  }

  @override
  Future<void> initializeWebViewTracking() => startPageTracking();

  @override
  Set<ContentsquareJavascriptChannel> get javascriptChannels {
    return ContentsquareJavascriptEventEnum.values
        .map(
          (event) => a147(
            a57: Contentsquare(),
            a56: event,
            a55: a71,
          ).a1,
        )
        .toSet();
  }

  
  String _a217({
    required String Function(String channelName, String data)
        a483,
  }) {
    return '''
window.CSJavascriptBridge = {
  
  ${ContentsquareJavascriptEventEnum.values.map(
              (event) => '${event.name}: ${a147(
                a57: Contentsquare(),
                a56: event,
                a55: a71,
              ).a20(
                a428: a483,
              )}',
            ).join(',\n\n').replaceAll('\n', '\n  ')}
};''';
  }

  
  Future<bool> _a218({
    required Future<bool> Function(String) a484,
  }) async {
    for (final a316 in javascriptChannels) {
      final a317 =
          await a484(a316.name);

      if (!a317) {
        return false;
      }
    }

    return true;
  }

  
  
  
  
  var _a318 = IList<Stopwatch>();

  
  Stopwatch _a219() {
    final a319 = Stopwatch();
    _a318 = _a318.add(a319);
    return a319;
  }

  
  void _a220(Stopwatch a485) {
    a485.stop();
    _a318 = _a318.remove(a485);
  }

  
  @visibleForTesting
  static String get a5 => '''
    if (window.cs_wvt == undefined || window.cs_wvt == null) {
      // Cast on 'null' for jsonDecode to decode it to dart null.
      'null';
    } else {
      window.cs_wvt.push(['serializeWebView']);
    }
  ''';

  
  Future<Map<String, dynamic>> a23() async {
    final a320 = _a312;
    if (a320 == null) {
      throw a176();
    }

    
    
    final a321 = _a219();
    a321.start();

    bool a221() {
      return a321.elapsed >=
          a72.a66;
    }

    
    
    
    
    bool a222() {
      return !a321.isRunning;
    }

    Map<String, dynamic>? a322;
    
    
    
    
    while (!(a221() || a222())) {
      final a323 = await a320.evaluateJavascript(
        a5,
      );
      final Object? a324;
      try {
        final a325 = jsonDecode(a323);
        if (a325 is String) {
          
          
          
          a324 = jsonDecode(a325);
        } else {
          a324 = a325;
        }
      } on FormatException catch (e) {
        a187.a39(
          a172(
            a42: a323,
            a41: e,
          ),
        );
        continue;
      }

      
      
      
      
      
      
      
      final a326 = a324 == null ||
          a324 is num ||
          a324 is Map<String, dynamic>;
      if (!a326) {
        a187.a39(
          a171(
            a43: a324,
          ),
        );
      }

      if (a324 is Map<String, dynamic>) {
        a322 = a324;
        break;
      }

      await Future.delayed(
        a72.a65,
      );
    }

    _a220(a321);

    if (a322 == null) {
      throw a173(
        a40: a72.a66,
      );
    }

    return a322;
  }

  
  void _a223() {
    _a318.forEach(_a220);
  }
}


class _a138 extends InheritedWidget {
  const _a138({
    Key? a486,
    required Widget a487,
  }) : super(key: a486, child: a487);

  @override
  bool updateShouldNotify(oldWidget) => false;
}


extension a182 on Element {
  
  bool get _a224 {
    return widget is ContentsquareWebViewTrackerBuilder;
  }

  
  bool get a225 {
    final a327 =
        getElementForInheritedWidgetOfExactType<_a138>();
    return a327 != null || _a224;
  }
}
