import 'dart:convert';

import 'package:contentsquare/src/modules/webview_tracking/javascript_event_handler.dart';
import 'package:contentsquare/src/modules/webview_tracking/logs/logs.dart';
import 'package:contentsquare_platform_interface/contentsquare_platform_interface.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mocktail/mocktail.dart';

import '../../../helpers/expect_logs.dart';
import '../../../helpers/mocks/mock_contentsquare.dart';

const fakeEncodedTVP = 'tvp';

main() {
  group(
    '${ContentsquareJavascriptEventEnum.sendEvent}',
    () {
      test(
        'should log a $WebViewTVPNotFoundErrorLog '
        'when no webview tracker TVP has been found',
        () async {
          expectLogs((log) => expect(log, isA<WebViewTVPNotFoundErrorLog>()));

          final _mockContentsquare = MockContentsquare();

          final _event = {
            'eventJson': jsonEncode({
              'type': ContentsquareGesture.pageView.toInt(),
              'data': {},
            })
          };

          final eventHandler = ContentsquareJavascriptEventHandler(
            contentsquare: _mockContentsquare,
            contentsquareEvent: ContentsquareJavascriptEventEnum.sendEvent,
            getEncodedWebViewTrackerTVP: () => null,
          );

          final handlerChannel = eventHandler.channel;

          handlerChannel.onMessageReceived(jsonEncode(_event));
        },
      );

      group('Gesture type is PageView', () {
        test(
          'should log a $InvalidJson when the url is not provided',
          () async {
            expectLogs((log) => expect(log, isA<InvalidJson>()));

            final _mockContentsquare = MockContentsquare();

            final _event = {
              'eventJson': jsonEncode({
                'type': ContentsquareGesture.pageView.toInt(),
                'data': {},
              })
            };

            final eventHandler = ContentsquareJavascriptEventHandler(
              contentsquare: _mockContentsquare,
              contentsquareEvent: ContentsquareJavascriptEventEnum.sendEvent,
              getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
            );

            final handlerChannel = eventHandler.channel;

            handlerChannel.onMessageReceived(jsonEncode(_event));
          },
        );

        test(
          'should call send method of Dart SDK '
          'when send page view event is sent',
          () async {
            const fakeURL = 'https://www.google.com';

            final _mockContentsquare = MockContentsquare();

            void flutterSDKSend() {
              _mockContentsquare.send(any());
            }

            when(flutterSDKSend).thenAnswer((_) => Future.value());

            final _event = {
              'eventJson': jsonEncode({
                'type': ContentsquareGesture.pageView.toInt(),
                'data': {'url': fakeURL},
              })
            };

            final eventHandler = ContentsquareJavascriptEventHandler(
              contentsquare: _mockContentsquare,
              contentsquareEvent: ContentsquareJavascriptEventEnum.sendEvent,
              getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
            );

            final handlerChannel = eventHandler.channel;

            handlerChannel.onMessageReceived(jsonEncode(_event));

            verify(flutterSDKSend).called(1);
          },
        );
      });

      group('Gesture type is tap or longPress', () {
        test(
          'should log a $InvalidJson when eventJson is invalid',
          () async {
            expectLogs((log) => expect(log, isA<InvalidJson>()));

            final _mockContentsquare = MockContentsquare();

            final eventHandler = ContentsquareJavascriptEventHandler(
              contentsquare: _mockContentsquare,
              contentsquareEvent: ContentsquareJavascriptEventEnum.sendEvent,
              getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
            );

            final handlerChannel = eventHandler.channel;

            handlerChannel.onMessageReceived('{"eventJson": "invalid" }');
          },
        );

        test(
          'should log a $InvalidJson when type is invalid',
          () async {
            expectLogs((log) => expect(log, isA<InvalidJson>()));

            const fakeVersion = 1;
            const fakePath = 'eq>:0:>div';
            final _mockContentsquare = MockContentsquare();

            final _event = {
              'eventJson': jsonEncode({
                'type': 'invalid',
                'version': fakeVersion,
                'data': {'path': fakePath},
              })
            };

            final eventHandler = ContentsquareJavascriptEventHandler(
              contentsquare: _mockContentsquare,
              contentsquareEvent: ContentsquareJavascriptEventEnum.sendEvent,
              getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
            );

            final handlerChannel = eventHandler.channel;

            handlerChannel.onMessageReceived(jsonEncode(_event));
          },
        );

        test(
          'should log a $InvalidJson when version is invalid ',
          () async {
            expectLogs((log) => expect(log, isA<InvalidJson>()));

            const fakeGesture = ContentsquareGesture.tap;
            const fakePath = 'eq>:0:>div';
            final _mockContentsquare = MockContentsquare();

            final _event = {
              'eventJson': jsonEncode({
                'type': fakeGesture.toInt(),
                'version': 'invalid',
                'data': {'path': fakePath},
              })
            };

            final eventHandler = ContentsquareJavascriptEventHandler(
              contentsquare: _mockContentsquare,
              contentsquareEvent: ContentsquareJavascriptEventEnum.sendEvent,
              getEncodedWebViewTrackerTVP: () {
                return '';
              },
            );

            final handlerChannel = eventHandler.channel;

            handlerChannel.onMessageReceived(jsonEncode(_event));
          },
        );

        test(
          'should log a $InvalidJson when path is invalid ',
          () async {
            expectLogs((log) => expect(log, isA<InvalidJson>()));

            const fakeGesture = ContentsquareGesture.tap;
            const fakeVersion = 1;
            final _mockContentsquare = MockContentsquare();

            final _event = {
              'eventJson': jsonEncode({
                'type': fakeGesture.toInt(),
                'version': fakeVersion,
                'data': {'path': 10},
              })
            };

            final eventHandler = ContentsquareJavascriptEventHandler(
              contentsquare: _mockContentsquare,
              contentsquareEvent: ContentsquareJavascriptEventEnum.sendEvent,
              getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
            );

            final handlerChannel = eventHandler.channel;

            handlerChannel.onMessageReceived(jsonEncode(_event));
          },
        );

        test(
          'should call sendTouchEvent method of Dart SDK '
          'when touch event is sent',
          () async {
            const fakePath = 'eq>:0:>div';
            const fakeGesture = ContentsquareGesture.tap;
            const fakeVersion = 1;

            final _mockContentsquare = MockContentsquare();

            void flutterSDKSendTouchEvent() {
              _mockContentsquare.sendTouchEvent(
                data: const TouchEventData(
                  path: fakeEncodedTVP + '|webview|' + fakePath,
                ),
                type: fakeGesture,
                version: fakeVersion,
              );
            }

            when(flutterSDKSendTouchEvent).thenAnswer((_) => Future.value());

            final _event = {
              'eventJson': jsonEncode({
                'type': fakeGesture.toInt(),
                'version': fakeVersion,
                'data': {'path': fakePath},
              })
            };

            final eventHandler = ContentsquareJavascriptEventHandler(
              contentsquare: _mockContentsquare,
              contentsquareEvent: ContentsquareJavascriptEventEnum.sendEvent,
              getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
            );

            final handlerChannel = eventHandler.channel;

            handlerChannel.onMessageReceived(jsonEncode(_event));

            verify(flutterSDKSendTouchEvent).called(1);
          },
        );
      });
    },
  );

  group('${ContentsquareJavascriptEventEnum.sendException}', () {
    test(
      'should log a $SendExceptionLog when the event is SendException ',
      () async {
        expectLogs((log) => expect(log, isA<SendExceptionLog>()));

        final _mockContentsquare = MockContentsquare();

        final eventHandler = ContentsquareJavascriptEventHandler(
          contentsquare: _mockContentsquare,
          contentsquareEvent: ContentsquareJavascriptEventEnum.sendException,
          getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
        );

        final handlerChannel = eventHandler.channel;

        handlerChannel.onMessageReceived(
          jsonEncode({'exception': 'an exception happened'}),
        );
      },
    );

    test(
      'should log a $InvalidJson when the event is invalid',
      () async {
        expectLogs((log) => expect(log, isA<InvalidJson>()));

        final _mockContentsquare = MockContentsquare();

        final eventHandler = ContentsquareJavascriptEventHandler(
          contentsquare: _mockContentsquare,
          contentsquareEvent: ContentsquareJavascriptEventEnum.sendException,
          getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
        );

        final handlerChannel = eventHandler.channel;

        handlerChannel.onMessageReceived(jsonEncode({'invalid': 'invalid'}));
      },
    );
  });

  group('${ContentsquareJavascriptEventEnum.sendMessage}', () {
    test(
      'should log a $SendMessageInfoLog when the event is SendMessage',
      () async {
        expectLogs((log) => expect(log, isA<SendMessageInfoLog>()));

        final _mockContentsquare = MockContentsquare();

        final eventHandler = ContentsquareJavascriptEventHandler(
          contentsquare: _mockContentsquare,
          contentsquareEvent: ContentsquareJavascriptEventEnum.sendMessage,
          getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
        );

        final handlerChannel = eventHandler.channel;

        handlerChannel.onMessageReceived(
          jsonEncode({'message': 'an incredible message'}),
        );
      },
    );

    test(
      'should log a $InvalidJson when the event is invalid',
      () async {
        expectLogs((log) => expect(log, isA<InvalidJson>()));

        final _mockContentsquare = MockContentsquare();

        final eventHandler = ContentsquareJavascriptEventHandler(
          contentsquare: _mockContentsquare,
          contentsquareEvent: ContentsquareJavascriptEventEnum.sendMessage,
          getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
        );

        final handlerChannel = eventHandler.channel;

        handlerChannel.onMessageReceived(jsonEncode({'invalid': 'invalid'}));
      },
    );
  });

  group('${ContentsquareJavascriptEventEnum.optIn}', () {
    test('should call optIn method of Dart SDK when OptIn event is sent',
        () async {
      final _mockContentsquare = MockContentsquare();

      void flutterSDKOptIn() => _mockContentsquare.optIn();

      when(flutterSDKOptIn).thenAnswer((_) => Future.value());

      final eventHandler = ContentsquareJavascriptEventHandler(
        contentsquare: _mockContentsquare,
        contentsquareEvent: ContentsquareJavascriptEventEnum.optIn,
        getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
      );

      final handlerChannel = eventHandler.channel;

      handlerChannel.onMessageReceived('{}');

      verify(flutterSDKOptIn).called(1);
    });
  });

  group('${ContentsquareJavascriptEventEnum.optOut}', () {
    test('should call optOut method of Dart SDK when OptOut event is sent',
        () async {
      final _mockContentsquare = MockContentsquare();

      void flutterSDKOptOut() => _mockContentsquare.optOut();

      when(flutterSDKOptOut).thenAnswer((_) => Future.value());

      final eventHandler = ContentsquareJavascriptEventHandler(
        contentsquare: _mockContentsquare,
        contentsquareEvent: ContentsquareJavascriptEventEnum.optOut,
        getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
      );

      final handlerChannel = eventHandler.channel;

      handlerChannel.onMessageReceived('{}');

      verify(flutterSDKOptOut).called(1);
    });
  });

  group('${ContentsquareJavascriptEventEnum.sendTransaction}', () {
    test(
      'should log a $InvalidJson when the json is invalid ',
      () async {
        expectLogs((log) => expect(log, isA<InvalidJson>()));

        final _mockContentsquare = MockContentsquare();

        final eventHandler = ContentsquareJavascriptEventHandler(
          contentsquare: _mockContentsquare,
          contentsquareEvent: ContentsquareJavascriptEventEnum.sendTransaction,
          getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
        );

        final handlerChannel = eventHandler.channel;

        handlerChannel.onMessageReceived('{}');
      },
    );
    test(
      'should call sendTransaction method of Dart SDK '
      'when SendTransaction event is sent',
      () async {
        const fakeId = 'my-id';
        const fakeValue = 100.0;
        const fakeCurrency = 'EUR';

        final _mockContentsquare = MockContentsquare();

        void flutterSDKSendTransaction() {
          _mockContentsquare.sendTransaction(
            fakeValue,
            fakeCurrency,
            fakeId,
          );
        }

        when(flutterSDKSendTransaction).thenAnswer((_) => Future.value());

        final _event = jsonEncode({
          'id': fakeId,
          'currency': fakeCurrency,
          'value': fakeValue,
        });

        final eventHandler = ContentsquareJavascriptEventHandler(
          contentsquare: _mockContentsquare,
          contentsquareEvent: ContentsquareJavascriptEventEnum.sendTransaction,
          getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
        );

        final handlerChannel = eventHandler.channel;

        handlerChannel.onMessageReceived(_event);

        verify(flutterSDKSendTransaction).called(1);
      },
    );

    test(
      'should call sendTransaction method of Dart SDK '
      'when SendTransaction event is sent without an id',
      () async {
        const fakeValue = 100.0;
        const fakeCurrency = 'EUR';

        final _mockContentsquare = MockContentsquare();

        void flutterSDKSendTransaction() {
          _mockContentsquare.sendTransaction(
            fakeValue,
            fakeCurrency,
          );
        }

        when(flutterSDKSendTransaction).thenAnswer((_) => Future.value());

        final _event = jsonEncode({
          'currency': fakeCurrency,
          'value': fakeValue,
        });

        final eventHandler = ContentsquareJavascriptEventHandler(
          contentsquare: _mockContentsquare,
          contentsquareEvent: ContentsquareJavascriptEventEnum.sendTransaction,
          getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
        );

        final handlerChannel = eventHandler.channel;

        handlerChannel.onMessageReceived(_event);

        verify(flutterSDKSendTransaction).called(1);
      },
    );
  });

  group('${ContentsquareJavascriptEventEnum.sendDynamicVar}', () {
    test(
      'should log a $InvalidJson when the json is invalid',
      () async {
        expectLogs((log) => expect(log, isA<InvalidJson>()));

        final _mockContentsquare = MockContentsquare();

        final eventHandler = ContentsquareJavascriptEventHandler(
          contentsquare: _mockContentsquare,
          contentsquareEvent: ContentsquareJavascriptEventEnum.sendDynamicVar,
          getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
        );

        final handlerChannel = eventHandler.channel;

        handlerChannel.onMessageReceived('{}');
      },
    );
    test(
      'should call sendDynamicVar method of Dart SDK '
      'when SendDynamic event is sent with a string value',
      () async {
        const fakeKey = 'my-key';
        const fakeStringValue = 'my-value';

        final _mockContentsquare = MockContentsquare();

        void flutterSDKSendDynamicVar() {
          _mockContentsquare.sendDynamicVar(
            fakeKey,
            stringValue: fakeStringValue,
          );
        }

        when(flutterSDKSendDynamicVar).thenAnswer((_) => Future.value());

        final _event = jsonEncode({
          'key': fakeKey,
          'value': fakeStringValue,
        });

        final eventHandler = ContentsquareJavascriptEventHandler(
          contentsquare: _mockContentsquare,
          contentsquareEvent: ContentsquareJavascriptEventEnum.sendDynamicVar,
          getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
        );

        final handlerChannel = eventHandler.channel;

        handlerChannel.onMessageReceived(_event);

        verify(flutterSDKSendDynamicVar).called(1);
      },
    );

    test(
      'should call sendDynamicVar method of Dart SDK '
      'when SendDynamic event is sent with an int value',
      () async {
        const fakeKey = 'my-key';
        const fakeIntValue = 1;

        final _mockContentsquare = MockContentsquare();

        void flutterSDKSendDynamicVar() {
          _mockContentsquare.sendDynamicVar(
            fakeKey,
            intValue: fakeIntValue,
          );
        }

        when(flutterSDKSendDynamicVar).thenAnswer((_) => Future.value());

        final _event = jsonEncode({
          'key': fakeKey,
          'value': fakeIntValue,
        });

        final eventHandler = ContentsquareJavascriptEventHandler(
          contentsquare: _mockContentsquare,
          contentsquareEvent: ContentsquareJavascriptEventEnum.sendDynamicVar,
          getEncodedWebViewTrackerTVP: () => fakeEncodedTVP,
        );

        final handlerChannel = eventHandler.channel;

        handlerChannel.onMessageReceived(_event);

        verify(flutterSDKSendDynamicVar).called(1);
      },
    );
  });
}
